/* --------------------------------------------------------------------------
 * Copyright 1992-1993 by Forschungszentrum Informatik (FZI)
 *
 * You can use and distribute this software under the terms of the licence
 * version 1 you should have received along with this software.
 * If not or if you want additional information, write to
 * Forschungszentrum Informatik, "STONE", Haid-und-Neu-Strasse 10-14,
 * D-76131 Karlsruhe, Germany.
 * --------------------------------------------------------------------------
 */
/* --------------------------------------------------------------------------
 * Arrows support routines
 *
 * ORIGINAL: Ch. Popp					Date: 2/6/93
 * --------------------------------------------------------------------------
 */
#include <stdio.h>
#include "tcl.h"
#include "arrows_tcl.h"


static double intersect (edgeX,edgeY,vectorX,vectorY,centerX,centerY)

   double edgeX,edgeY,vectorX,vectorY,centerX,centerY;

{
 double a1 = (edgeX - centerX) / vectorX,
        a2 = (edgeY - centerY) / vectorY;

 if (a1 < 0) a1 = -a1;
 if (a2 < 0) a2 = -a2;

 return (a1 > a2) ? a2 : a1;
}


/* -------
 * Compute coordinates for the 'continuous' movement mode.
 */
static edge_coords_continuous (cd, interp, argc, argv)

   ClientData	cd;
   Tcl_Interp	*interp;
   int		argc;
   char		*argv[];
 
{ 
 double startX  = atoi(argv[1]),
 	startY  = atoi(argv[2]),
 	destX   = atoi(argv[3]),
 	destY   = atoi(argv[4]),
 	widthS  = atoi(argv[5]),
 	widthD  = atoi(argv[6]),
 	heightS = atoi(argv[7]),
 	heightD = atoi(argv[8]);

 int    edgestartX, edgestartY, edgedestX, edgedestY;

 double deltaY  = startY - destY,
 	deltaX  = startX - destX,
 	a, edgeX, edgeY;
 char*  position;

 if (deltaX == 0) {
    edgestartX = startX;
    edgedestX  = startX;

    if (deltaY >= 0) {
       edgestartY = startY - (heightS/2);
       edgedestY  = destY  + (heightD/2);
       position   = "NE";
    } else {
       edgestartY = startY + (heightS/2);
       edgedestY  = destY  - (heightD/2);
       position   = "SE";
    }

 } else if (deltaY == 0) {
    edgestartY = startY;
    edgedestY  = startY;

    if (deltaX >= 0) {
       edgestartX = startX - (widthS/2);
       edgedestX  = destX  + (widthD/2);
       position   = "SW";
     } else {
       edgestartX = startX + (widthS/2);
       edgedestX  = destX  - (widthD/2);
       position   = "NW";
     }

 } else if (deltaX < 0) {
    if (deltaY >= 0) {
       /* 1. quadrant */

       edgeX      = startX + (widthS/2);
       edgeY      = startY - (heightS/2);
       a          = intersect(edgeX,edgeY,deltaX,deltaY,startX,startY);
       edgestartX = startX - (a * deltaX);
       edgestartY = startY - (a * deltaY);

       edgeX     = destX - (widthD/2);
       edgeY     = destY + (heightD/2);
       a         = intersect(edgeX,edgeY,deltaX,deltaY,destX,destY);
       edgedestX = destX + (a * deltaX);
       edgedestY = destY + (a * deltaY);

       position  = "NE";

    } else {
       /* 4. quadrant */

       edgeX      = startX + (widthS/2);
       edgeY      = startY + (heightS/2);
       a          = intersect(edgeX,edgeY,deltaX,deltaY,startX,startY);
       edgestartX = startX - (a * deltaX);
       edgestartY = startY - (a * deltaY);

       edgeX     = destX - (widthD/2);
       edgeY     = destY - (heightD/2);
       a         = intersect(edgeX,edgeY,deltaX,deltaY,destX,destY);
       edgedestX = destX + (a * deltaX);
       edgedestY = destY + (a * deltaY);

       position = "SE";
    }
 } else if (deltaY < 0) {
    /* 3. quadrant */

    edgeX      = startX - (widthS/2);
    edgeY      = startY + (heightS/2);
    a          = intersect(edgeX,edgeY,deltaX,deltaY,startX,startY);
    edgestartX = startX - (a * deltaX);
    edgestartY = startY - (a * deltaY);

    edgeX     = destX + (widthD/2);
    edgeY     = destY - (heightD/2);
    a         = intersect(edgeX,edgeY,deltaX,deltaY,destX,destY);
    edgedestX = destX + (a * deltaX);
    edgedestY = destY + (a * deltaY);

    position = "SW";

 } else {
    /* 2. quadrant */

    edgeX      = startX - (widthS/2);
    edgeY      = startY - (heightS/2);
    a          = intersect(edgeX,edgeY,deltaX,deltaY,startX,startY);
    edgestartX = startX - (a * deltaX);
    edgestartY = startY - (a * deltaY);

    edgeX     = destX + (widthD/2);
    edgeY     = destY + (heightD/2);
    a         = intersect(edgeX,edgeY,deltaX,deltaY,destX,destY);
    edgedestX = destX + (a * deltaX);
    edgedestY = destY + (a * deltaY);

    position = "NW";
 }
	/* we have at least 200 bytes - and that's enough */
 sprintf (interp->result,
	  "%s %d %d %d %d",
	  position, edgestartX, edgestartY, edgedestX, edgedestY);

 return TCL_OK;
}


/* -------
 * Compute coordinates for the 'fast' movement mode.
 */
static edge_coords_fast (cd, interp, argc, argv)

   ClientData	cd;
   Tcl_Interp	*interp;
   int		argc;
   char		*argv[];
 
{ 
 int startX  = atoi(argv[1]),
 	startY  = atoi(argv[2]),
 	destX   = atoi(argv[3]),
 	destY   = atoi(argv[4]),
 	widthS  = atoi(argv[5]),
 	widthD  = atoi(argv[6]),
 	heightS = atoi(argv[7]),
 	heightD = atoi(argv[8]);

 int    edgestartX, edgestartY, edgedestX, edgedestY;
 char*  position;

 int    start_HighY  = startY - (heightS/2),
        start_LowY   = startY + (heightS/2),
 	start_RightX = startX + (widthS/2),
 	start_LeftX  = startX - (widthS/2),
 	dest_HighY   = destY  - (heightD/2),
 	dest_LowY    = destY  + (heightD/2),
 	dest_RightX  = destX  + (widthD/2),
 	dest_LeftX   = destX  - (widthD/2);

 if (start_LowY <= dest_HighY) {
    edgestartY = start_LowY;
    edgestartX = startX;
    edgedestY  = dest_HighY;
    edgedestX  = destX;
    position   = "S";

 } else if (start_HighY >= dest_LowY) {
    edgestartY = start_HighY;
    edgestartX = startX;
    edgedestY  = dest_LowY;
    edgedestX  = destX;
    position   = "N";

 } else if (start_RightX < dest_LeftX) {
    edgestartY = startY;
    edgestartX = start_RightX;
    edgedestY  = destY;
    edgedestX  = dest_LeftX;
    position   = "E";

 } else {
    edgestartY = startY;
    edgestartX = start_LeftX;
    edgedestY  = destY;
    edgedestX  = dest_RightX;
    position   = "W";
 }
	/* we have at least 200 bytes - and that's enough */
 sprintf (interp->result,
	  "%s %d %d %d %d",
	  position, edgestartX, edgestartY, edgedestX, edgedestY);

 return TCL_OK;
}

/* -------
 * module initialization
 */
void arrow_init (interp)

   Tcl_Interp* interp;

{
  Tcl_CreateCommand(interp,
		    "arrow_coords_continuous_C",
		    (Tcl_CmdProc*)edge_coords_continuous,
		    (ClientData)NULL,(Tcl_CmdDeleteProc*)NULL);
  Tcl_CreateCommand(interp,
		    "arrow_coords_fast_C",
		    (Tcl_CmdProc*)edge_coords_fast,
		    (ClientData)NULL,(Tcl_CmdDeleteProc*)NULL);
}
